#include <u.h>
#include <libc.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include "layer.h"
#include "dat.h"
#include "fns.h"

enum {
	Chunk = 1024,
	Ulbits = 8*sizeof(ulong),
};
typedef struct Intbitmap Intbitmap;
struct Intbitmap {
	QLock;
	ulong	*free;		/* ptr to a free bit, if any */
	ulong	offset;
	ulong	used[Chunk/Ulbits+1];
	ulong	*eused;
	Intbitmap *link;
};

struct Intpool {
	QLock;
	ulong	n;		/* allocate 0..n-1 */
	Intbitmap *head;
	Intbitmap *tail;
	Intbitmap *free;	/* earliest bitmap in chain with free tags */
};

static Intbitmap*
allocbitmap(ulong base, ulong max)
{
	Intbitmap *m;
	ulong *p;
	ulong l;

	m = mallocz(sizeof *m, 1);
	if (m == nil)
		return nil;

	m->eused = &m->used[nelem(m->used)-1];	/* sentinel */
	*m->eused = 0;
	m->free = m->used;
	m->offset = base;

	if (base+Chunk > max) {
		max -= base;
		/* mark as used all bits >= max */
		p = &m->used[max/Ulbits];
		if (max%Ulbits) {
			for (l = 1<<(max%Ulbits); l; l <<= 1)
				*p |= l;
			p++;
		}
		m->eused = p;
		*p = 0;
	}
	return m;
}

ulong
allocint(Intpool *pool)
{
	int i;
	ulong l, rv, b, *p;
	Intbitmap *m;

	qlock(pool);
	for (m = pool->free; m; m = m->link)
		if (m->free)
			break;

	/* allocate new bitmap if necessary */
	if (m == nil) {
		if (pool->tail) {
			if (pool->tail->offset+Chunk >= pool->n) {
				qunlock(pool);
				return ~0UL;
			}
			m = allocbitmap(pool->tail->offset+Chunk, pool->n);
			if (m == nil) {
				qunlock(pool);
				return ~0UL;
			}
			pool->tail->link = m;
		} else {
			m = allocbitmap(0, pool->n);
			if (m == nil) {
				qunlock(pool);
				return ~0UL;
			}
			pool->head = m;
		}
		pool->tail = pool->free = m;
	}

	assert(m->used <= m->free && m->free < m->eused);
	/* take bit */
	b = *m->free;
	for (i = 0, l = 1; b&l; i++, l <<= 1)
		;
	assert(l != 0);
	*m->free |= l;
	rv = m->offset + (m->free - m->used)*Ulbits + i;
	/* advance free pointer */
	for (p = m->free; *p == ~0UL; p++)
		;

	if (p == m->eused)
		m->free = nil;
	else
		m->free = p;
	qunlock(pool);
	return rv;
}

void
freeint(Intpool *pool, ulong ul)
{
	Intbitmap *m;
	ulong *p;

	qlock(pool);
	assert(ul < pool->n);

	/* find correct bitmap */
	if (pool->free && pool->free->offset <= ul)
		m = pool->free;
	else
		m = pool->head;

	for (; m && m->offset + Chunk <= ul; m = m->link)
		;
	assert(m != nil);
	assert(m->offset <= ul && ul < m->offset+Chunk);
	ul -= m->offset;

	/* mark as not used */
	p = &m->used[ul/Ulbits];
	assert(*p & (1<<(ul%Ulbits)));
	*p &= ~(1<<(ul%Ulbits));

	/* update bitmap free ptr */
	if (m->free == nil || p < m->free)
		m->free = p;

	/* update pool free ptr */
	if (pool->free == nil || pool->free->offset > m->offset)
		pool->free = m;
	qunlock(pool);
}

Intpool*
allocintpool(ulong n)
{
	Intpool *p;

	p = mallocz(sizeof *p, 1);
	if (p != nil)
		p->n = n;
	return p;
}

void
freeintpool(Intpool *p)
{
	Intbitmap *m, *mnext;

	qlock(p);
	for (m = p->head; m; m = mnext) {
		mnext = m->link;
		free(m);
	}
	qunlock(p);
	free(p);
}
