/*
 * talk to the file servers using a distinct fid space for each session
 */
#include <u.h>
#include <libc.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include "layer.h"
#include "dat.h"
#include "fns.h"

// send 9P request *r over s->reqchan, receive response into *r via nonce chan
int
sendrecv9p(Session *s, Req *r, char *e)
{
	int type;
	Req *nr;

	assert(r->buf == nil);
	assert(r->magic == Reqmagic);
	r->c = chancreate(sizeof r, 0);
	type = r->type;
	// TODO: post r somewhere so that it can be flushed 
	sendp(s->reqchan, r);

	nr = recvp(r->c);
	// TODO: unpost r
	chanfree(r->c);
	assert(nr == r);
	assert(r->magic == Reqmagic);

	if (r->type == type+1) {
		if (r->type != Rread)
			nnfree(&r->buf);
		return 0;
	} else if (r->type == Rerror)
		strecpy(e, e+ERRLEN, r->ename);
	else
		strncpy(e, "mismatched T & R messages", ERRLEN);
	return -1;
}

/*
 * recv 9P requests on v->rdchan, unmap tag→r for session,
 * send r's buf & Fcall on via r's c.
 */
void
read9pthread(void *v)
{
	Req *r, *or;
	Session *s = v;

	while ((r = recvp(s->rdchan)) != nil) {
		assert(r->magic == Reqmagic);
		or = deletekey(s->tagmap, r->tag);	// unmap tag→r
		if (or == nil)
			sysfatal("unexpected reply %d", r->tag);
		unlockmap(s->tagmap);
		assert(or->magic == Reqmagic);
		or->buf = r->buf;
		or->Fcall = r->Fcall;
		r->buf = nil;			// don't free or->buf
		donereq(r);
		sendp(or->c, or);
	}
	threadexits(nil);
}

/*
 * recv 9P requests on v->reqchan, map tag→r for session, convert r's Fcall to
 * machine-independent form and write it down s->fd.
 */
void
write9pthread(void *v)
{
	int n;
	char buf[MAXMSG+MAXFDATA];
	Req *r;
	Session *s = v;

	while ((r = recvp(s->reqchan)) != nil) {
		assert(r->magic == Reqmagic);
		assert(r->buf == nil);
		r->tag = (r->type == Tnop || r->type == Tsession? NOTAG:
			allocint(s->tagpool));
		if (chatty)
			threadprint(2, "write %s %F\n", s->name, &r->Fcall);
		assert(caninsertkey(s->tagmap, r->tag, r) == 0);
		if ((n = convS2M(&r->Fcall, buf)) <= 0)
			sysfatal("9p write bad message type %d", r->type);
		if (write9p(s->fd, buf, n) != n)
			sysfatal("9p write error on fd %d: %r", s->fd);
	}
	threadexits(nil);
}

/*
 * initialise a Session matching the file server connection on fd.
 * start the necessary processes and threads.
 */
Session *
initsession(int fd, int fsno, char *name)
{
	char e[ERRLEN];
	Readarg *ra;
	Session *s;

	s = emalloc(sizeof *s);
	s->fd = fd;
	s->fsno = fsno;
	s->reqchan = chancreate(sizeof(Req*), 0); /* buffer for efficiency? */
	s->rdchan = chancreate(sizeof(Req*), 0);  /* buffer for efficiency? */
	s->fidpool = allocintpool(Maxfid);  // our fid space for this session
	s->tagpool = allocintpool(Maxtag);  // our tag space for this session
	s->tagmap = allocmap();
	s->name = name;

	// process #2 or #3
	ra = emalloc(sizeof *ra);
	ra->fd = s->fd;
	ra->c = s->rdchan;
	ra->name = name;
	proccreate(read9pproc, ra, Readstack);		// process #2 or #3

	threadcreate(read9pthread, s, Rdwrstack);	// thread #4.2 or 4.4
	threadcreate(write9pthread, s, Rdwrstack);	// thread #4.3 or 4.5

	if (txsession(s, e) < 0)			// authenticate to s->fd
		sysfatal("authentication to %s failed", name);
	return s;
}

static void
resetmeta(Fid *f)
{
	memset(&f->Fidmeta, 0, sizeof f->Fidmeta);
	incref(&f->ref);
}

static Fid *
allocfid(Session *s)
{
	Fid *f;

	f = emallocz(sizeof *f, 1);
	f->s = s;
	f->fid = allocint(s->fidpool);
	f->qid = (Qid){0, 0};
	f->magic = Fidmagic;
	resetmeta(f);
	return f;
}

static void
freefid(Fid *f)
{
#ifdef PARANOID
	assert(canqlock(&f->flock));
	qunlock(&f->flock);
#endif
	freeint(f->s->fidpool, f->fid);
	f->magic = 0;		// decertify
	free(f);
}

void
donefid(Fid *f)
{
	assert(f->magic == Fidmagic);
	if (decref(&f->ref) == 0)
		freefid(f);
}

// lower Fid's reference count; if it reaches zero, clunk & free the Fid.
void
fidfrclunk(Fid **f)
{
	char e[ERRLEN];

	if (*f == nil)
		return;
	assert((*f)->magic == Fidmagic);
	if (decref(&(*f)->ref) == 0) {
		fidclunk(*f, e);
		freefid(*f);
		*f = nil;
	}
}

Req *
newreq(Fid *fid, int type)
{
	Req *r = emalloc(sizeof *r);

	incref(r);			// caller holds a ref
	r->magic = Reqmagic;
	r->type = type;
	r->buf = nil;
	if (fid != nil) {
		assert(fid->magic == Fidmagic);
		r->fid = fid->fid;
	}
	return r;
}

/* drop one reference to r; if none left, free r */
void
donereq(Req *r)
{
	if (r == nil)
		return;
	assert(r->magic == Reqmagic);
	if (decref(r) == 0) {
		r->magic = 0;		// decertify
		r->c = nil;
		nnfree(&r->buf);
		free(r);
	}
}

// exchange attach messages, including authentication
Fid *
fidattach(Session *s, char *uid, char *aname, Qid *qid, char *e)
{
	Fid *fid = nil;
	Req *rp = newreq(nil, Tattach);

	strecpy(rp->uname, rp->uname+NAMELEN, uid);
	strecpy(rp->aname, rp->aname+NAMELEN, aname);
	if (setticket(s, &rp->Fcall, e) >= 0) {
		fid = allocfid(s);
		rp->fid = fid->fid;

		if (sendrecv9p(s, rp, e) < 0) {
			donefid(fid);
			fid = nil;
		} else if (checkattach(s, &rp->Fcall, e) < 0) {
			fidclunk(fid, rp->ename);		// BUG?
			fid = nil;
		} else
			*qid = fid->qid = rp->qid;
	}
	donereq(rp);
	return fid;
}

// fid must not yet have been opened via open nor create.
Fid *
fidclone(Fid *fid, char *e)
{
	Fid *newfid;
	Req *rp = newreq(fid, Tclone);

	assert(fid != nil);
	newfid = allocfid(fid->s);
	rp->newfid = newfid->fid;
	if (sendrecv9p(fid->s, rp, e) < 0) {
		donefid(newfid);
		newfid = nil;
	} else
		newfid->qid = fid->qid;
	donereq(rp);
	return newfid;
}

static int
doitqid(Req *rp, Fid *fid, Qid *qid, char *e)
{
	int rv = sendrecv9p(fid->s, rp, e);

	if (qid != nil && rv >= 0)
		*qid = fid->qid = rp->qid;
	donereq(rp);
	return rv;
}

/*
 * only walk & create change the meaning of a fid.
 */
int
fidwalk(Fid *fid, char *name, Qid *qid, char *e)
{
	Req *rp = newreq(fid, Twalk);

	strecpy(rp->name, rp->name + sizeof rp->name, name);
	return doitqid(rp, fid, qid, e);
}

int
fidopen(Fid *fid, int mode, Qid *qid, char *e)
{
	Req *rp = newreq(fid, Topen);

	rp->mode = mode;
	return doitqid(rp, fid, qid, e);
}

/*
 * only walk & create change the meaning of a fid.
 */
int
fidcreate(Fid *fid, char *name, int mode, ulong perm, Qid *qid, char *e)
{
	Req *rp = newreq(fid, Tcreate);

	rp->mode = mode;
	rp->perm = perm;
	strecpy(rp->name, rp->name+NAMELEN, name);
	return doitqid(rp, fid, qid, e);
}

long
fidread(Fid *fid, void *data, long count, vlong offset, char *e)
{
	Req *rp = newreq(fid, Tread);

	rp->count = count;
	rp->offset = offset;
	if (sendrecv9p(fid->s, rp, e) < 0)
		count = -1;
	else {
		memmove(data, rp->data, rp->count);
		count = rp->count;
	}
	donereq(rp);
	return count;
}

long
fidwrite(Fid *fid, void *data, long count, vlong offset, char *e)
{
	Req *rp = newreq(fid, Twrite);

	rp->data = data;
	rp->count = count;
	rp->offset = offset;
	count = (sendrecv9p(fid->s, rp, e) < 0? -1: rp->count);
	donereq(rp);
	return count;
}

// frees fid after clunking it
static int
clunkremove(Fid *fid, int type, char *e)
{
	Req *rp = newreq(fid, type);
	int rv = sendrecv9p(fid->s, rp, e);

	donefid(fid);
	donereq(rp);
	return rv;
}

// frees fid after clunking it
int
fidclunk(Fid *fid, char *e)
{
	return clunkremove(fid, Tclunk, e);
}

// frees fid after clunking it
int
fidremove(Fid *fid, char *e)
{
	return clunkremove(fid, Tremove, e);
}

int
fidstat(Fid *fid, char *stat, char *e)
{
	Req *rp = newreq(fid, Tstat);
	int rv = sendrecv9p(fid->s, rp, e);

	if (rv >= 0)
		memmove(stat, rp->stat, DIRLEN);
	donereq(rp);
	return rv;
}

int
fidwstat(Fid *fid, char *stat, char *e)
{
	Req *rp = newreq(fid, Twstat);

	memmove(rp->stat, stat, DIRLEN);
	return doitqid(rp, fid, nil, e);
}

int
fidflush(Fid *fid, ushort oldtag, char *e)
{
	Req *rp = newreq(fid, Tflush);

	rp->oldtag = oldtag;
	return doitqid(rp, fid, nil, e);
}
