/*
 * the externally-visible routines in this file lock internally
 * to protect Map structures against concurrent access.
 */
#include <u.h>
#include <libc.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include "layer.h"
#include "dat.h"
#include "fns.h"

enum {
	Mapmagic = 0xdeadbeef,
};

typedef struct Maplist Maplist;
struct Maplist {
	ulong	key;
	void	*val;
	Maplist	*link;
};

static Maplist **
lfind(Maplist **l, ulong key)
{
	for (; *l; l = &(*l)->link)
		if ((*l)->key == key)
			break;
	return l;
}

Map *
allocmap(void)
{
	Map *map = mallocz(sizeof(Map), 1);

	if (map != nil)
		map->magic = Mapmagic;
	return map;
}

void
freemap(Map *m)
{
	Maplist *l, *lnext;

	assert(m->magic == Mapmagic);
	qlock(m);
	for (l = m->list; l; l = lnext) {
		lnext = l->link;
		free(l);
	}
	m->magic = 0;
	qunlock(m);
	free(m);
}

/*
 * non-nil return value leaves m locked.  why?  to keep (*l)->val from being
 * deallocated while we manipulate it?
 */
void *
lookupkey(Map *m, ulong key)
{
	Maplist **l;

	assert(m->magic == Mapmagic);
	qlock(m);
	l = lfind(&m->list, key);
	if (*l == nil) {
		qunlock(m);
		return nil;
	} else
		return (*l)->val;
}

void *
insertkey(Map *m, ulong key, void *val)
{
	Maplist **l;
	void *rv;

	assert(m->magic == Mapmagic);
	qlock(m);
	l = lfind(&m->list, key);
	if (*l == nil) {
		*l = emallocz(sizeof **l, 1);
		(*l)->key = key;
		(*l)->val = val;
		qunlock(m);
		return nil;		// wasn't found, but was inserted
	}
	qunlock(m);
	rv = (*l)->val;
	(*l)->val = val;
	return rv;			// was found; changed value
}

int
caninsertkey(Map *m, ulong key, void *val)
{
	Maplist **l;

	assert(m->magic == Mapmagic);
	qlock(m);
	l = lfind(&m->list, key);
	if (*l == nil) {
		*l = emallocz(sizeof **l, 1);
		(*l)->key = key;
		(*l)->val = val;
		qunlock(m);
		return 0;		// wasn't found, but was inserted
	}
	qunlock(m);
	return -1;			// was found; no change
}

static Maplist **
prevfind(Maplist **l, ulong key)
{
	Maplist **prev = nil;

	for (; *l; prev = l, l = &(*l)->link)
		if ((*l)->key == key)
			break;
	return prev;
}

// if result is non-nil, leave m locked
void *
deletekey(Map *m, ulong key)
{
	Maplist **prev, *this;
	void *rv;

	assert(m->magic == Mapmagic);
	qlock(m);
	prev = prevfind(&m->list, key);
	if (prev != nil && (*prev == nil || (*prev)->link == nil)) { // GOK
		qunlock(m);
		return nil;
	}
	if (prev == nil && m->list != nil && m->list->key == key) // we're 1st?
		this = m->list;
	else if (prev != nil)
		this = (*prev)->link;
	else {
		qunlock(m);
		return nil;
	}
	if (this->key != key) {		// prevfind is buggered?
		qunlock(m);
		return nil;
	}

	// unlink this
	if (prev == nil && m->list == this)
		m->list = this->link;	// was no prior Maplist
	else
		(*prev)->link = this->link;
	this->link = nil;
	// no longer: qunlock(m);

	rv = this->val;
	free(this);
	return rv;
}

void
unlockmap(Map *m)
{
	qunlock(m);
}
