#include <u.h>
#include <libc.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include "layer.h"
#include "dat.h"
#include "fns.h"

#define MAX(a, b) ((a) > (b)? (a): (b))
#define MIN(a, b) ((a) < (b)? (a): (b))
#define rounddown(n, div) (((n)/(div))*(div))

static char Efidinuse[] = "fid already in use";

/* from /sys/src/9/^*^/errstr.h */
// char Enoerror[] = "no error";
// char Emount[] = "inconsistent mount";
// char Eunmount[] = "not mounted";
// char Eunion[] = "not in union";
// char Emountrpc[] = "mount rpc error";
// char Eshutdown[] = "mounted device shut down";
// char Enocreate[] = "mounted directory forbids creation";
char Enonexist[] = "file does not exist";
// char Eexist[] = "file already exists";
// char Ebadsharp[] = "unknown device in # filename";
char Enotdir[] = "not a directory";
// char Eisdir[] = "file is a directory";
// char Ebadchar[] = "bad character in file name";
// char Efilename[] = "file name syntax";
// char Eperm[] = "permission denied";
// char Ebadusefd[] = "inappropriate use of fd";
char Ebadarg[] = "bad arg in system call";
// char Einuse[] = "device or object already in use";
// char Eio[] = "i/o error";
// char Etoobig[] = "read or write too large";
// char Etoosmall[] = "read or write too small";
// char Enoport[] = "network port not available";
// char Ehungup[] = "write to hungup channel";
// char Ebadctl[] = "bad process or channel control request";
// char Enodev[] = "no free devices";
// char Eprocdied[] = "process exited";
// char Enochild[] = "no living children";
// char Eioload[] = "i/o error in demand load";
// char Enovmem[] = "virtual memory allocation failed";
// char Ebadfd[] = "fd out of range or not open";
// char Enofd[] = "no free file descriptors";
// char Eisstream[] = "seek on a stream";
// char Ebadexec[] = "exec header invalid";
// char Etimedout[] = "connection timed out";
// char Econrefused[] = "connection refused";
// char Eintr[] = "interrupted";
// char Enomem[] = "kernel allocate failed";
// char Enoswap[] = "swap space full";
// char Esoverlap[] = "segments overlap";
// char Emouseset[] = "mouse type already set";
// char Eshort[] = "i/o count too small";
// char Egreg[] = "xterm: Error 50, errno 1: Too big";
// char Ebadspec[] = "bad attach specifier";
// char Enoreg[] = "process has no saved registers";
// char Enoattach[] = "mount/attach disallowed";

enum {
	Ovfidleft,
	Ovfidfreed,

	Ovfidmagic = 0xdeadbabe,
};

/*
 * one of our stitched (overlaid) fids.
 * only copy Ovfdata, not Ovfmeta, when copying Ovfids.
 * these correspond to client fids; the client allocates
 * those fids and is responsible for avoiding fid races,
 * so there shouldn't be races accessing them.  the allocator
 * locks the Ovfid pool internally.
 */
typedef struct {
#ifdef PARANOID
	QLock	oflock;
#endif
	Ref	ref;
} Ovfmeta;
typedef struct {
	ulong	magic;
	int	fidno;		/* in our client's fid space */
	int	numfids;
	Fid	*fids[Maxfs];	/* nil: no file at that position in this fs */
	Fid	*dupfids[Maxfs]; /* for dir.s, dup these into fids */
	int	fsno;		/* if non-negative, fids index for this file */
	/*
	 * we need the name so we can walk .. in all the Fids of an Ovfid,
	 * even the nil ones.
	 */
	char	*name;		// full name if exists, else potential full name
} Ovfdata;
typedef struct {
	Ovfmeta;
	Ovfdata;
} Ovfid;

typedef struct {
	Req	*r;
	Map	*ctagmap;
	int	numsess;
	Session *sess[Maxfs];

	Map	*fidmap;		// our client's fid space
} Threadarg;

static Ovfid *rootovfid;

static void stitchfn(void*);

static Session *
initsess(int srvfd, int fsno, char *name)
{
	Session *sess = initsession(srvfd, fsno, name);

	if (sess == nil)
		sysfatal("initsession: %s: %r", name);
	return sess;
}

/* the 9P request launcher, eventually; first launch other threads */
void
stitchthread(void *v)
{
	int fs, rd9ppid, wr9ptid;
	Channel *creply, *csrv;
	Readarg *ra, *wa;
	Req *r;
	Stitcharg *a = v;
	Threadarg *ta, tmp;

threadprint(2, "ovfs: main process started\n");
	rfork(RFFDG /* |RFNAMEG|RFNOTEG */ );
	close(a->otherpipeend);		// seal off the other end of the pipe
	a->otherpipeend = -1;
threadprint(2, "ovfs: main process id is %d\n", getpid());

	tmp.numsess = a->numfs;
	tmp.ctagmap = allocmap();	// tags our client uses
	for (fs = 0; fs < a->numfs; fs++)
		tmp.sess[fs] = initsess(a->fs[fs].fsfd, fs, a->fs[fs].fsname);

	csrv =   chancreate(sizeof(Req*), 0);
	creply = chancreate(sizeof(Req*), 0);

	// process #1 sends us Reqs from the served namespace
	ra = emalloc(sizeof *ra);
	ra->fd = a->srvfd;
	ra->c = csrv;
	ra->name = "stitch←";
	rd9ppid = proccreate(read9pproc, ra, Readstack);	// process #1
	assert(rd9ppid > 0);

	// thread #X: receives Reqs from us and writes to the served namespace
	wa = emalloc(sizeof *wa);
	wa->fd = a->srvfd;
	wa->c = creply;
	wa->name = "stitch→";
	wr9ptid = threadcreate(write9pproc, wa, Readstack);	// thread #X
	assert(wr9ptid > 0);

	/*
	 * thread #4.1: 9P request launcher
	 *
	 * why is it safe to process each incoming 9P request in a separate
	 * thread?  Russ Cox (rsc) explains:
	 *
	 * The client gets no ordering guarantees, but 9P doesn't make any.
	 * If the client wants something to happen in order it has to send
	 * one, wait for the response, send the next.
	 *
	 *   I thought I remembered reading that 9P relied on the underlying
	 *   transport (IL or TCP) for in-order delivery.
	 *
	 * This much is true.  But it's required so that you get flush
	 * right, not for general operation.  Of course, ovfs doesn't get
	 * flush right, and in-order answers there would fix that.
	 */
	tmp.fidmap = allocmap();
	while ((r = recvp(csrv)) != nil) {
		assert(r->magic == Reqmagic);
		incref(r);		// tmp.ctagmap will hold a ref
		if (insertkey(tmp.ctagmap, r->tag, r) != nil)
			threadprint(2, "tag %d in use; bug in client\n",r->tag);
		r->c = creply;
		ta = emalloc(sizeof *ta);
		*ta = tmp;
		ta->r = r;
		if (singleth)
			stitchfn(ta);
		else {
			// thread #4.x
			int tid = threadcreate(stitchfn, ta, Reqstack);

			assert(tid > 0);
//			threadprint(2,
//			    "stitchthread: started work thread %d for Req\n",
//				tid);
		}
	}
threadprint(2, "ovfs: main process (and all threads) exiting\n");
	yield();
	sleep(50);
	threadexitsall("eof on request channel");
}

static void
seterr(Req *r, char *err)
{
	r->type = Rerror;
	if (err != nil)
		strncpy(r->ename, err, ERRLEN);
}

static void
diagmismatch(Req *r, Ovfid *of, int fs)
{
	char err[ERRLEN];

	if (fs > 0 && of->fids[fs-1]->qid.path&CHDIR)
		snprint(err, sizeof err, "directory over plain file in fs%d",
			fs);
	else
		snprint(err, sizeof err, "plain file over directory in fs%d",
			fs);
	seterr(r, err);
}

// returns fs # of first real (non-nil) fid in `of', if any; else -1
static int
firstfid(Ovfid *of)
{
	int fs;

	assert(of->magic == Ovfidmagic);
	for (fs = 0; fs < of->numfids && of->fids[fs] == nil; fs++)
		continue;
	if (fs >= of->numfids)
		return -1;
	assert(of->fids[fs]->magic == Fidmagic);
	return fs;
}

/*
 * return CHDIR or 0 iff all files in `of' are of the same type
 * (directory or plain file, resp.), else fs # of first difference
 */
static int
firstdifftype(Ovfid *of)
{
	int fs = firstfid(of), type;

	if (fs < 0)
		return 0;		// slight lie
	type = of->fids[fs]->qid.path&CHDIR;
	for (; fs < of->numfids; fs++) {
		Fid *fid = of->fids[fs];

		if (fid != nil && ((type ^ fid->qid.path) & CHDIR) != 0)
			return fs;
	}
	return type;
}

static Fid *
curfid(Ovfid *of)
{
	int fs = of->fsno;

	assert(of->magic == Ovfidmagic);
	if (fs < 0 || of->fids[fs] == nil) {
		fs = firstfid(of);
	// TODO: how can this happen?  what are failure semantics of Twalk?
		assert(fs >= 0 /* ovfid has a fid */);
	}
	assert(of->fids[fs] != nil);
	assert(of->fids[fs]->magic == Fidmagic);
	return of->fids[fs];
}

/*
 * trivial authentication to this program, relying on authentication
 * to underlying file servers?
 */
static void
stitchsession(Req *r)
{
	memset(r->chal, 0, sizeof r->chal);
	memset(r->authid, 0, sizeof r->authid);
	memset(r->authdom, 0, sizeof r->authdom);
}

/* map our client's fid # to a locked Ovfid */
static Ovfid *
fidno2ovfid(Req *r, int fidno, Map *fidmap)
{
	char err[ERRLEN];
	Ovfid *ovfid = lookupkey(fidmap, fidno);

	if (ovfid == nil) {
		snprint(err, sizeof err, "unknown fid %d", fidno);
		seterr(r, err);
	} else {
		assert(ovfid->magic == Ovfidmagic);
		assert(ovfid->fidno == fidno);
#ifdef PARANOID
		while (!canqlock(&ovfid->oflock)) {
			threadprint(2, "stitchfn: waiting for Ovfid lock\n");
			sleep(5000);
		}
#endif
		// non-nil lookupkey() value leaves map locked
		unlockmap(fidmap);
	}
	return ovfid;
}

/*
 * abort any operations with our client's tag r->oldtag.
 * it's not easy to interrupt the relatively synchronous operations
 * being conducted by the fid* routines called from the stitch* routines.
 * it's important that slow reads be interruptible, though.
 */
static void
stitchflush(Req *r, Threadarg *ta)
{
	Req *oldr;

	// map r->oldtag to oldr, then ovfid, thence Reqs & tags for our servers
	oldr = lookupkey(ta->ctagmap, r->oldtag);
	if (oldr == nil)
		return;			// r->oldtag not in use by our client
	// ctagmap is locked, so deletemap can't occur until unlockmap
	// so oldr must be valid.
	assert(oldr->magic == Reqmagic);
	assert(oldr->Ref.ref > 0);	// else lookupkey couldn't have found it
	oldr->flushed = 1;
	unlockmap(ta->ctagmap);  // non-nil lookupkey() value leaves map locked
	/*
	 * TODO? Tflush in-flight requests to file servers on behalf of
	 * oldr(->tag).  we need some locking against the process(es) doing
	 * the work that we're trying to abort.
	 */
}

static void
resetmeta(Ovfid *of)
{
	memset(&of->Ovfmeta, 0, sizeof of->Ovfmeta);
	incref(&of->ref);
}

static int
freeovfid(Ovfid *of)
{
	assert(of->magic == Ovfidmagic);
	if (decref(&of->ref) != 0)
		return Ovfidleft;
	if (chatty)	// DEBUG
		threadprint(2, "freeovfid freeing\n");
#ifdef notdef
	assert(canqlock(&of->oflock));
	qunlock(&of->oflock);
#endif
	nnfree(&of->name);
	of->magic = 0;		// decertify
	resetmeta(of);
	free(of);
	return Ovfidfreed;
}

static void
offidclunk(Fid **fidp)
{
	fidfrclunk(fidp);
}

static void
offidsclunk(Ovfid *of, int fs)
{
	offidclunk(&of->fids[fs]);
	offidclunk(&of->dupfids[fs]);
}

// clunks fids from 0 to `fs-1'
static void
ofclunk(Ovfid *of, int fs)
{
	assert(of->magic == Ovfidmagic);
	while (--fs >= 0)
		offidsclunk(of, fs);
}

// clunks fids from 0 to `fs-1' & frees ovfid.
static int
offrclunk(Ovfid *of, int fs)
{
	ofclunk(of, fs);
	return freeovfid(of);
}

// ugh.  put fs # in high qid path bits to distinguish.
static Qid
qidmash(Qid qid, int fs)
{
	Qid mash = qid;

	mash.path |= (fs&((1<<Fsbits)-1)) << (32-Rsvbits-Fsbits);
	return mash;
}

static int
qidunmash(Qid mash, Qid *qidp)
{
	Qid qid = mash;
	int fs = Qidfs(qid.path);

	qid.path &= ~(((1<<Fsbits)-1) << (32-Rsvbits-Fsbits));
	*qidp = qid;
	return fs;
}

// must attach to all file servers
static void
stitchattach(Req *r, Map *fidmap, int numsess, Session *sess[])
{
	int fs;
	char err[ERRLEN];
	Fcall or;
	Qid q;
	Ovfid *of;

	of = rootovfid = emallocz(sizeof *of, 1);
	of->magic = Ovfidmagic;
	of->fsno = -1;
	resetmeta(of);

	or = r->Fcall;
	of->numfids = numsess;
	for (fs = 0; fs < numsess; fs++) {
		of->fids[fs] = fidattach(sess[fs], or.uname, or.aname, &q, err);
		if (of->fids[fs] == nil) {
			(void) offrclunk(of, fs);
			seterr(r, err);
			return;
		}
		if (fs == 0)
			r->qid = qidmash(q, 0);
	}

	of->fidno = r->fid;		// client's fid of the served root
	if (caninsertkey(fidmap, r->fid, of) < 0) {
		(void) offrclunk(of, fs);
		seterr(r, Efidinuse);
	} else {
		of->name = estrdup("/");
		memset(r->rauth, 0, sizeof r->rauth);
	}
}

static void
incfidrefs(Ovfid *of)
{
	int fs;

	for (fs = 0; fs < of->numfids; fs++) {
		if (of->fids[fs] != nil)
			incref(&of->fids[fs]->ref);
		if (of->dupfids[fs] != nil)
			incref(&of->dupfids[fs]->ref);
	}
}

static void
ofdcopy(Ovfid *nof, Ovfid *of)
{
	assert(of->magic == Ovfidmagic);
	nof->Ovfdata = of->Ovfdata;
	nof->name = estrdup(of->name);
	incfidrefs(of);
}

// the duplicate will not be locked, which is fine, since no one else will
// have a reference to it yet.  it will also have a reference count of 1.
static Ovfid *
ofdup(Ovfid *of)
{
	Ovfid *nof = emalloc(sizeof *nof);

	assert(of->magic == Ovfidmagic);
	resetmeta(nof);
	ofdcopy(nof, of);
	return nof;
}

static Ovfid *
ofclone(Ovfid *of, char *err)
{
	int fs;

	for (fs = 0; fs < of->numfids; fs++) {
		Fid *fid = of->fids[fs];

		if (fid == nil)
			continue;
		fidfrclunk(&of->fids[fs]);
		of->fids[fs] = fid = fidclone(fid, err);
		if (fid == nil) {
			offrclunk(of, fs);
			return nil;
		}
	}
	return of;
}

/*
 * clone any non-nil Fids in of, per r->fid & r->newfid.
 * clone all in case r->name is a directory, thus possibly a union.
 * only walk & create change the meaning of a fid.
 * Ovfid must not yet have been opened via open nor create.
 */
static int
stitchclone(Req *r, Map *fidmap, Ovfid *of)
{
	char err[ERRLEN];
	Ovfid *nof = ofclone(ofdup(of), err);	// nof is not locked

	assert(firstfid(of) >= 0);
	if (nof == nil) {
		seterr(r, err);
		return Ovfidleft;
	}
	assert(firstfid(nof) >= 0);
	nof->fidno = r->newfid;
	if (caninsertkey(fidmap, r->newfid, nof) < 0) {
		offrclunk(nof, nof->numfids);
		snprint(err, sizeof err, "%s: %d", Efidinuse, r->newfid);
		seterr(r, err);
	}
	return Ovfidleft;
}

/*
 * returned value is malloced.
 */
static char *
walkname(char *dir, char *file)
{
	int dotdot = strcmp(file, "..") == 0;

	assert(dir != nil && dir[0] == '/');
	if (file[0] == '\0' || strcmp(file, ".") == 0 ||
	    dotdot && strcmp(dir, "/") == 0)
		return strdup(dir);		// no movement
	if (dotdot) {				// walk up?
		char *rs;

		dir = strdup(dir);
		rs = utfrrune(dir, '/');
		if (rs == dir)
			dir[1] = '\0';
		else
			*rs = '\0';
	} else if (strcmp(dir, "/") == 0) {	// walk down from root?
		dir = emalloc(1 + strlen(file) + 1);
		dir[0] = '/';
		strcpy(dir + 1, file);
	} else {				// walk down from non-root
		char *origdir = dir;

		dir = emalloc(strlen(dir) + 1 + strlen(file) + 1);
		strcpy(dir, origdir);
		strcat(dir, "/");
		strcat(dir, file);
	}
	return dir;
}

/* return value is malloced; dir must be too */
static char *
updname(char *dir, char *comp)
{
	char *origdir = dir;

	dir = walkname(dir, comp);
	free(origdir);
	return dir;
}

/*
 * returned value is malloced and it is assumed that `dir' is too.
 * the returned value may be a modified version of dir.
 */
static char *
wstatname(char *dir, char *file)
{
	return updname(updname(dir, ".."), file);
}

static Qid
fidopenmash(Req *r, Fid *fid, int mode, char *err)
{
	Qid qid = (Qid){ 0, 0 };

	if (fid == nil)
		return qid;
	assert(fid->magic == Fidmagic);
	if (fidopen(fid, mode, &qid, err) < 0)
		seterr(r, nil);
	else
		qid = qidmash(qid, fid->s->fsno);
	return qid;
}

static int ofwalk(Req *r, char *name, Ovfid *of, Qid *qidp, char *, int);

static void
ofwalkfromroot(Req *r, Ovfid *of, Qid *qidp, char *err, int fsstart)
{
	int ncomp;
	char *parent = walkname(of->name, "..");
	char *sarray[64];
	char **compp;
	Ovfid *wof = ofclone(ofdup(rootovfid), err);	// wof is not locked

	ncomp = getfields(parent, sarray, nelem(sarray), 1, "/");
	if (ncomp == nelem(sarray))
		threadprint(2, "more than %d components in a full path\n",
			nelem(sarray));
	for (compp = sarray; compp < sarray + ncomp && r->type != Rerror;
	     compp++)
		ofwalk(r, *compp, wof, qidp, err, fsstart);
	ofdcopy(of, wof);
	offrclunk(wof, wof->numfids);
	free(parent);
}

/*
 * walk to all in case name is a directory, thus possibly a union.
 * only walk & create change the meaning of a (ov)fid.
 */
static int
ofwalk(Req *r, char *name, Ovfid *of, Qid *qidp, char *err, int fsstart)
{
	int dotdot, fs, fromroot = 0;
	Qid qid1;
	Fid *fid = nil;

	// we have to walk all the fids in `of' to stay in step
	assert(firstfid(of) >= 0);
	dotdot = strcmp(name, "..") == 0;
	for (fs = fsstart; fs < of->numfids; fs++) {
		fid = of->fids[fs];
		if (fid == nil)
			continue;
		assert(fid->magic == Fidmagic);
		if (fidwalk(fid, name, &qid1, err) == 0) {
			if (of->fsno < 0) {	// note uppermost
				of->fsno = fs;
				*qidp = qid1;
				if (!(qid1.path&CHDIR)) { // plain file; done
					fidfrclunk(&of->dupfids[fs]);
					while (++fs < of->numfids)
						offidsclunk(of, fs);
					break;
				}
			}
		} else {
			offidsclunk(of, fs);	// nobody home in this fs
			if (dotdot) {
				fromroot = 1;
				break;
			}
		}
	}
	if (fs < of->numfids && fromroot) {
		threadprint(2, "warning: walk .. on fs %s failed\n",
			fid->s->name);
		ofwalkfromroot(r, of, qidp, err, fsstart);
		return 1;
	}
	if (firstfid(of) < 0) {
		if (of->numfids <= 0)
			threadprint(2, "ofwalk: no fids in result\n");
		strncpy(err, Enonexist, ERRLEN);
		seterr(r, nil);
	}
	return 0;
}

static int
stitchwalk(Req *r, Ovfid *of)
{
	int fromroot;
	char err[ERRLEN];
	Qid qid;
	Fid *fid;
	Ovfid *nof;

	assert(firstfid(of) >= 0);
	fid = of->fids[firstfid(of)];
	if (fid == nil) {
		seterr(r, Enonexist);
		return Ovfidleft;
	}
	assert(fid->magic == Fidmagic);
	if (!(fid->qid.path & CHDIR)) {	// plain file; can't walk into it
		seterr(r, Enotdir);
		return Ovfidleft;
	}
	if (strcmp(r->name, ".") == 0) {
		r->qid = qidmash(fid->qid, fid->s->fsno);
		return Ovfidleft;
	}

	err[0] = '\0';
	nof = ofclone(ofdup(of), err);	// nof is not locked
	if (nof == nil) {
		seterr(r, err);
		return Ovfidleft;
	}
	assert(firstfid(nof) >= 0);
	nof->fsno = -1;

	// walk all fids in `nof'
	fromroot = ofwalk(r, r->name, nof, &qid, err, 0);
	if (nof->fsno < 0) {		// failed: not even 1 file named r->name
		threadprint(2, "stitchwalk: failed to walk `%s' in `%s': %s\n",
			r->name, of->name, err);	// DEBUG
		seterr(r, err);
	} else {
		assert(firstfid(nof) >= 0);
		ofclunk(of, of->numfids);	// success: clunk old Ovfid
		ofdcopy(of, nof);		// nof is not locked
		r->qid = qidmash(qid, of->fsno);
		if (!fromroot)
			of->name = updname(of->name, r->name);
	}
	offrclunk(nof, nof->numfids);  // if failed, implicit clunk of new Ovfid
	return Ovfidleft;
}

static int
stitchclwalk(Req *r, Map *fidmap, Ovfid *of)
{
	if (stitchclone(r, fidmap, of) == Ovfidfreed)
		return Ovfidfreed;
	if (r->type != Rerror)
		// implicit clunk of (new) `of' on failure of stitchwalk
		return stitchwalk(r, of);
}

/*
 * If top file is a directory, we open all directories.
 * Otherwise, just open the current one and clunk the others.
 */
static int
stitchopen(Req *r, Ovfid *of)
{
	int fs, mode, type, failed;
	char err[ERRLEN];
	Qid qid, retqid;
	Fid *fid;

	err[0] = '\0';
	// save a few variables so they won't be overwritten when r->ename
	// is set.
	mode = r->mode;
	type = r->type;
	assert(firstfid(of) >= 0);
	fs = firstfid(of);
	if (fs < 0) {			/* no fids? */
		seterr(r, Enonexist);
		return Ovfidleft;
	}
	fid = of->fids[fs];
	assert(fid->magic == Fidmagic);
	if (fid->qid.path&CHDIR) {
//		int firstdiff = firstdifftype(of);
//		if (firstdiff > 0) {
//			diagmismatch(r, of, firstdiff);
//			return Ovfidleft;
//		}
		for (; fs < of->numfids; fs++) {
			fid = of->fids[fs];
			if (fid == nil)
				continue;
			assert(fid->magic == Fidmagic);
			if (!(fid->qid.path&CHDIR)) {
				offidsclunk(of, fs);
				continue;
			}
			// clone before opening so we can clone later
			fidfrclunk(&of->dupfids[fs]);
			of->dupfids[fs] = fidclone(fid, err);
			if (of->dupfids[fs] == nil)
				continue;
			r->type = type;
			qid = fidopenmash(r, fid, mode, err);
			if (r->type != Rerror && fs == 0)
				retqid = qid;
		}
		failed = (firstfid(of) < 0);

	} else {
		retqid = fidopenmash(r, fid, mode, err);
		failed = (r->type == Rerror);
		while (++fs < of->numfids)
			if (of->fids[fs] != nil)
				offidsclunk(of, fs);
	}
	if (failed)
		seterr(r, err);
	else
		r->qid = retqid;
	return Ovfidleft;
}

// create in the top-most layer, at least for now.  walk the others.
static int
stitchcreate(Req *r, Ovfid *of)
{
	char err[ERRLEN];

	// assume that of->fids[0] will be unchanged if fidcreate fails.
	if (fidcreate(of->fids[0], r->name, r->mode, r->perm, &r->qid, err) < 0)
		seterr(r, err);
	else {
		Req *othr = newreq(of->fids[0], r->type);

		r->qid = qidmash(r->qid, 0);
		of->fsno = 0;
		// walk other fids in `nof'
		if (!ofwalk(othr, r->name, of, &othr->qid, err, 1))
			// didn't have to walk from root, which fixes of->name
			of->name = updname(of->name, r->name);
		if (firstfid(of) < 0)
			seterr(r, err);
		donereq(othr);
	}
	return Ovfidleft;
}

static void
statqidmash(char *stat, int fs)
{
	Dir dir;

	convM2D(stat, &dir);
	dir.qid = qidmash(dir.qid, fs);
	convD2M(&dir, stat);
}


typedef struct {
	vlong	cdoff;		// into the current directory
	vlong	unoff;		// into the union directory
	char	*buf;
	long	count;		// bytes left to read
} Io;

#define nfilled(iop, r)	((iop)->buf - (r)->data)
#define nboring(iop, r) ((r)->offset - (iop)->unoff)
#define readmore(iop, after) ((iop)->count >= DIRLEN && (iop)->unoff < (after))

enum {
	Direrr = -1,
	Direof = -2,
};

static long
rddir(Req *r, Fid *fid, Io *iop, int fs, char *err)
{
	long count, nbytes;
	vlong afteroff = r->offset + r->count, aftercount;
	char *dent, *buf, *next;
	Dir dir;

	count = MIN(iop->count, MAXFDATA);
	// don't read uninteresting & interesting data into the
	// same buffer; stop just before interesting data.
	aftercount = iop->unoff + count;
	if (iop->unoff < r->offset && aftercount > r->offset)
		count = MIN(count, (long)nboring(iop, r));
	else if (iop->unoff >= r->offset && aftercount > afteroff)
		count = MIN(count, (long)(afteroff - iop->unoff));
	count = rounddown(count, DIRLEN); // must read multiples of DIRLEN bytes
	if (count == 0)
		return 0;		// no data, not EOF
	if (count < 0)
		return Direrr;

	buf = iop->buf;
	nbytes = fidread(fid, iop->buf, count, iop->cdoff, err);
	if (nbytes == 0)
		return Direof;
	if (nbytes < 0)
		return Direrr;

	/* if bytes are in range, mangle qids for client & accept the bytes */
	if (iop->unoff >= r->offset && iop->unoff < afteroff) {
		for (dent = iop->buf; dent < iop->buf + nbytes; dent += DIRLEN)
			statqidmash(dent, fs);
		iop->count -= nbytes;
		iop->buf += nbytes;
	}
	iop->cdoff += nbytes;
	iop->unoff += nbytes;

	/* paranoia: don't let . nor .. dir. entries through */
	for (dent = buf; dent < iop->buf; ) {
		next = dent + DIRLEN;
		convM2D(dent, &dir);
		if (dir.name[0] == '.' &&
		    (strcmp(dir.name, ".")  ==0 || strcmp(dir.name, "..")==0)) {
			memmove(dent, next, iop->buf - next);
			iop->count += DIRLEN;
			iop->buf -= DIRLEN;
			nbytes -= DIRLEN;
		} else
			dent = next;
	}
	return nbytes;
}

/*
 * If top and bottom are in sync and directories,
 * we read them in sequence.
 * Otherwise, just read from top.
 * one can't seek on a directory, so clone the fid each time.
 */
static void
rdunion(Req *r, Ovfid *of, int fsstart, char *err)
{
	int fs;
	long nbytes = Direof;
	vlong afteroff = r->offset + r->count;
	Io io;

	if (r->offset < 0 || r->count < 0 || r->count < DIRLEN) {
		seterr(r, Ebadarg);
		return;
	}
	/*
	 * read r->offset bytes in the concatentation of the directories in
	 * the union.  if that doesn't pass EOF, return up to r->count bytes.
	 */
	io.unoff = 0;
	io.count = r->count;
	io.buf = r->data;
	for (fs = fsstart; fs < of->numfids && nbytes == Direof &&
	    readmore(&io, afteroff); fs++) {
		Fid *fid = of->fids[fs], *cfid;
		Qid qid = (Qid){ 0, 0 };

		if (fid == nil || !(fid->qid.path&CHDIR) ||
		    of->dupfids[fs] == nil)
			continue;		// nobody home or not a dir.

		// dupfids are deliberately not yet opened, so open a clone
		cfid = fidclone(of->dupfids[fs], err);
		if (cfid == nil || fidopen(cfid, OREAD, &qid, err) < 0) {
			seterr(r, err);
			return;
		}

		io.cdoff = 0;			// new underlying dir.
		while (readmore(&io, afteroff))
			if ((nbytes = rddir(r, cfid, &io, fs, err)) < 0)
				break;

		(void) fidclunk(cfid, err);

		if (nbytes == Direof && fs >= of->numfids-1)
			break;			// final eof of the union
	}
	if (nbytes == Direrr)
		seterr(r, err);
	else
		r->count = nfilled(&io, r);	// zero = "end of union"
}

/*
 * fill r->data.
 */
static int
stitchread(Req *r, Ovfid *of)
{
	int fs;
	char err[ERRLEN];

	err[0] = '\0';
	if (r->buf == nil)
		r->buf = emalloc(MAXMSG+MAXFDATA);
	if (r->data == nil)
		r->data = r->buf;
	assert(firstfid(of) >= 0);
	fs = firstfid(of);
	if (fs >= 0 && of->fids[fs]->qid.path&CHDIR) {
		rdunion(r, of, fs, err);
		if (err[0] != '\0')
			seterr(r, err);
	} else if (fidread(curfid(of), r->data, r->count, r->offset, err) < 0)
		seterr(r, err);
	return Ovfidleft;
}

// write r->data
static int
stitchwrite(Req *r, Ovfid *of)
{
	if (r->data == nil)
		r->data = r->buf;
	else if (r->buf == nil)
		r->buf = r->data;
	assert(firstfid(of) >= 0);
	if (fidwrite(curfid(of), r->data, r->count, r->offset, r->ename) < 0)
		seterr(r, nil);
	return Ovfidleft;
}

static int
stitchclunk(Req *r, Map *fidmap, Ovfid *of)
{
	Ovfid *dof;

	USED(r);
	assert(firstfid(of) >= 0);
	dof = deletekey(fidmap, of->fidno);
	assert(dof != nil);
	unlockmap(fidmap);
	assert(dof == of);
	offrclunk(of, of->numfids);
	return Ovfidfreed;
}

// just remove top-most file
static int
stitchremove(Req *r, Map *fidmap, Ovfid *of)
{
	int fs = firstfid(of);

	assert(firstfid(of) >= 0);
	if (fidremove(of->fids[fs], r->ename) < 0) {
		seterr(r, nil);
		return Ovfidleft;
	}
	of->fids[fs] = nil;	// Tremove (& fidremove()) clunked it implicitly
	return stitchclunk(r, fidmap, of);
}

static int
stitchstat(Req *r, Ovfid *of)
{
	Fid *fid = curfid(of);

	if (fidstat(fid, r->stat, r->ename) < 0)
		seterr(r, nil);
	else
		statqidmash(r->stat, fid->s->fsno);
	return Ovfidleft;
}

static int
stitchwstat(Req *r, Ovfid *of)
{
	Dir dir;

	convM2D(r->stat, &dir);
	if (fidwstat(curfid(of), r->stat, r->ename) < 0)
		seterr(r, nil);
	else {
		char *origdir = of->name;

		of->name = wstatname(of->name, dir.name);
		free(origdir);
	}
	return Ovfidleft;
}

// thread #4.x: work for Req *r
static void
stitchfn(void *a)
{
	int dowhat = Ovfidfreed, type, tag;
	Threadarg *ta = a;
	Req *r = ta->r, *delr;
	Ovfid *ovfid;

//threadprint(2, "stitchfn: got T message %F\n", &r->Fcall); // DEBUG
	type = r->type;
	tag = r->tag;
	switch (type) {
	default:
		seterr(r, "unknown T message type");
		break;
	case Tnop:
		break;
	case Tsession:
		stitchsession(r);
		break;
	case Tattach:
		stitchattach(r, ta->fidmap, ta->numsess, ta->sess);
		break;
	case Tflush:
		stitchflush(r, ta);
		break;

	case Tclone:	case Tclwalk:
	case Tclunk:	case Tremove:
	case Twalk:
	case Topen:	case Tcreate:
	case Tread:	case Twrite:
	case Tstat:	case Twstat:
		/* map our client's fid to Ovfid */
		ovfid = fidno2ovfid(r, r->fid, ta->fidmap);
		if (ovfid == nil)	// errstr already set
			break;
		switch (type) {
		case Tclone:
			dowhat = stitchclone(r, ta->fidmap, ovfid);
			break;
		case Tclwalk:
			dowhat = stitchclwalk(r, ta->fidmap, ovfid);
			break;
		case Tclunk:
			dowhat = stitchclunk(r, ta->fidmap, ovfid);
			break;
		case Tremove:
			dowhat = stitchremove(r, ta->fidmap, ovfid);
			break;

		case Twalk:
			dowhat = stitchwalk(r, ovfid);
			break;
		case Topen:
			dowhat = stitchopen(r, ovfid);
			break;
		case Tcreate:
			dowhat = stitchcreate(r, ovfid);
			break;
		case Tread:
			dowhat = stitchread(r, ovfid);
			break;
		case Twrite:
			dowhat = stitchwrite(r, ovfid);
			break;
		case Tstat:
			dowhat = stitchstat(r, ovfid);
			break;
		case Twstat:
			dowhat = stitchwstat(r, ovfid);
			break;
		default:
			dowhat = Ovfidfreed;
			break;
		}
		switch (dowhat) {
		case Ovfidleft:
#ifdef PARANOID
			if (chatty)	// DEBUG
				threadprint(2, "releasing Ovfid lock");
			qunlock(&ovfid->oflock);
#endif
			break;
		case Ovfidfreed:
		default:
#ifdef PARANOID
			if (chatty)	// DEBUG
				threadprint(2, "not releasing Ovfid lock");
#endif
			break;
		}
#ifdef PARANOID
		if (chatty)	// DEBUG
			threadprint(2, " at request end\n");
#endif
		break;
	}
	USED(dowhat);
	if (r->type != Rerror)
		r->type++;

	// break the "tag # → Req *" mapping
	delr = deletekey(ta->ctagmap, tag);
	// ctagmap is locked iff delr != nil, so delr can't be re-inserted
	// until unlockmap
	if (delr == nil)
		threadprint(2,
"%s: warning: deletekey(ta->ctagmap, r->tag) == nil for r->tag=%d r->type=%d\n",
			argv0, tag, type);
	else if (delr != r)
		threadprint(2,
"%s: warning: deletekey(ta->ctagmap, r->tag) != r for delr=0x%lux r->tag=%d r->type=%d\n",
			argv0, (ulong)delr, tag, type);

	// toss the ref from ctagmap
	assert(r->magic == Reqmagic);
	if (delr != nil) {
		assert(delr->magic == Reqmagic);
		donereq(delr);
		unlockmap(ta->ctagmap);
	}
	assert(r->magic == Reqmagic);

	if (r->flushed)
		donereq(r);
	else {
//threadprint(2, "stitchfn: sent R message %F\n", &r->Fcall); // DEBUG
		sendp(r->c, r);		// receiving thread does donereq(r)
	}

	free(ta);
}
