/*
 * deal with layers of directory trees (plan 9 specific)
 */

#include "mk.h"

/*
 * on brazil, reading a single Dir at a time from /mnt/term/mnt/8½ failed.
 * ls reads 50 at a time, so do the same.
 */
enum { Hunk = 50 };

static int
issamefs(Dir *dp1, Dir *dp2)
{
	return dp1->dev == dp2->dev && dp1->type == dp2->type &&
		Qidfs(dp1->qid.path) == Qidfs(dp2->qid.path);
}

static int
issamefsinst(Fsinst *dp1, Fsinst *dp2)
{
	return dp1->dev == dp2->dev && dp1->type == dp2->type &&
		Qidfs(dp1->qid.path) == Qidfs(dp2->qid.path);
}

Layering *
newlayer(Fsinst fs)
{
	Layering *nlp = Malloc(sizeof *nlp);

	nlp->Fsinst = fs;
	nlp->lower = nil;
	return nlp;
}

static Layering *				/* malloced chain */
layers(char *dirname, int fd)
{
	int first = 1, n, dirnlen;
	char *name;
	Layering *hlp = nil, *tlp = nil, *nlp;
	Dir dents[Hunk], lastdir, *dentp;

	while ((n = dirread(fd, dents, sizeof dents)) > 0)
		for (n /= sizeof *dents, dentp = dents; --n >= 0; ++dentp) {
			if (first)
				first = 0;
			else if (issamefs(&lastdir, dentp))
				continue;
			/* change of layer */
			nlp = newlayer((Fsinst){dentp->type, dentp->dev,
				dentp->qid});

			/* add this fs to the tail of this dir's layers */
			if (hlp == nil)
				hlp = nlp;
			else
				tlp->lower = nlp;
			tlp = nlp;

			/* insert the file's metadata into the symbol table */
			dirnlen = strlen(dirname);
			name = Malloc(dirnlen + 1 + strlen(dentp->name) + 1);
			memmove(name, dirname, dirnlen);
			name[dirnlen] = '/';
			strcpy(name+dirnlen+1, dentp->name);
			symlook(name, S_LAYERS, (void *)nlp);
			symlook(name, S_TIME, (void*)dentp->mtime);

			lastdir = *dentp;
		}
	return hlp;
}

static void
freelayers(Layering *lp)
{
	Layering *next;

	for (; lp != nil; lp = next) {
		next = lp->lower;
		lp->lower = nil;
		free(lp);
	}
}

static char *					/* malloced */
dirname(char *file)
{
	char *s, *sl;

	if (file == nil || (sl = utfrrune(file, '/')) == nil)
		return strdup(".");
	else if (sl == file)
		return strdup("/");
	else {
		*sl = '\0';
		s = strdup(file);
		*sl = '/';
		return s;
	}
}

static Layering *				/* malloced */
dirlayers(char *name)
{
	int fd;
	Layering *lp;
	Symtab *sym;

	sym = symlook(name, S_LAYERS, 0);	/* Layering cached? */
	if (sym)
		return (Layering *)sym->value;

	fd = open(name, OREAD);
	if (fd < 0 /* || dirfstat(fd, &dir) < 0 */ )
		return nil;
	lp = layers(name, fd);
	close(fd);
	symlook(strdup(name), S_LAYERS, (void *)lp);	/* cache dir layers */
	return lp;
}

static void
dumpfs(char *dirn, Layering *lays)
{
	fprint(2, "mk: directory %s unioned with:\n", dirn);
	for (; lays != nil; lays = lays->lower)
		fprint(2, " #%C%d", lays->type, lays->dev);
	fprint(2, "\n");
}

/*
 * is target in a lower layer of binds than prereq?
 */
int
islowerfs(Node *targ, Node *prereq)
{
	int depth = 1, targdep = 0, preqdep = 0, ret = 0;
	char *targd, *prereqd;
	Layering *tlp;

	targd =   dirname(targ->name);
	prereqd = dirname(prereq->name);
	if (strcmp(targd, prereqd) != 0) {
		/*
		 * we know prereq->Fsinst will be found in dirlayers(prereqd),
		 * and targ->Fsinst in tlays.
		 */
		for (tlp = dirlayers(targd); tlp != nil;
		     tlp = tlp->lower, depth++) {
			if (issamefsinst(&tlp->Fsinst, &prereq->Fsinst))
				preqdep = depth;
			if (issamefsinst(&tlp->Fsinst, &targ->Fsinst))
				targdep = depth;
		}
		if (preqdep == 0) {
			if (explain)
				fprint(2,
"mk: warning: target %s and prereq %s directories' overlap doesn't matter\n",
					targ->name, prereq->name);
		} else
			ret = (targdep > preqdep);
	}
	free(targd);
	free(prereqd);
	return ret;
}
